import asyncio
import logging
from typing import Final, cast

import readchar
import typer
from serial import Serial

from smpmgr.common import Options

logger = logging.getLogger(__name__)

MAP_KEY_TO_BYTES: Final = {
    readchar.key.CTRL_C: b"\x03",
    readchar.key.UP: b"\x1b[A",
    readchar.key.DOWN: b"\x1b[B",
    readchar.key.LEFT: b"\x1b[D",
    readchar.key.RIGHT: b"\x1b[C",
    readchar.key.ESC: b"\x1b",
}


def terminal(ctx: typer.Context) -> None:
    """Open a terminal to the device."""

    options = cast(Options, ctx.obj)

    async def f() -> None:
        if options.transport.port is None:
            print("--port <port> option is required for the terminal, e.g.")
            print("smpmgr --port COM1 terminal")
            return

        print(f"\x1b[2mOpening terminal to {options.transport.port}...", end="")

        with Serial(port=options.transport.port, baudrate=115200, timeout=options.timeout) as s:
            print("OK")
            print("Press Ctrl-T to exit the terminal.\x1b[22m")
            print()
            device_result, keyboard_result = await asyncio.wait(
                (
                    asyncio.create_task(_rx_from_device(s)),
                    asyncio.create_task(asyncio.to_thread(_tx_keyboard_to_device, s)),
                ),
                return_when=asyncio.FIRST_EXCEPTION,
            )

            logger.debug(f"{device_result=}, {keyboard_result=}")

    asyncio.run(f())


async def _rx_from_device(port: Serial) -> None:
    """Async poll of the serial port for incoming data.

    Can be replaced when pyserial is replaced."""

    while True:
        _bytes = port.read_all()
        if _bytes is not None:
            print(_bytes.decode(), end="", flush=True)
        await asyncio.sleep(0.020)


def _tx_keyboard_to_device(port: Serial) -> None:
    """Blocking read of keyboard input."""
    while True:
        try:
            key = readchar.readkey()
        except KeyboardInterrupt:
            key = readchar.key.CTRL_C
        if key == readchar.key.CTRL_T:
            raise KeyboardInterrupt
        try:
            port.write(MAP_KEY_TO_BYTES[key])
        except KeyError:
            port.write(key.encode())



"""The image subcommand group."""

import asyncio
import logging
from io import BufferedReader
from pathlib import Path
from typing import cast

import typer
from rich import print
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from smp.exceptions import SMPBadStartDelimiter
from smpclient import SMPClient
from smpclient.generics import error, success
from smpclient.mcuboot import ImageInfo
from smpclient.requests.image_management import ImageStatesRead, ImageStatesWrite
from typing_extensions import Annotated

from smpmgr.common import Options, connect_with_spinner, get_smpclient, smp_request

app = typer.Typer(name="image", help="The SMP Image Management Group.")
logger = logging.getLogger(__name__)


@app.command()
def state_read(ctx: typer.Context) -> None:
    """Request to read the state of FW images on the SMP Server."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)

        r = await smp_request(smpclient, options, ImageStatesRead(), "Waiting for image states...")  # type: ignore # noqa

        if error(r):
            print(r)
        elif success(r):
            if len(r.images) == 0:
                print("No images on device!")
            for image in r.images:
                print(image)
            if r.splitStatus is not None:
                print(f"splitStatus: {r.splitStatus}")
        else:
            raise Exception("Unreachable")

    asyncio.run(f())


@app.command()
def state_write(
    ctx: typer.Context,
    hash: Annotated[str, typer.Argument(help="SHA256 hash of the image header and body.")],
    confirm: Annotated[bool, typer.Argument(help="Confirm the image given by hash.")],
) -> None:
    """Request to write the state of FW images on the SMP Server."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)
    hash_bytes = bytes.fromhex(hash)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)

        r = await smp_request(smpclient, options, ImageStatesWrite(hash=hash_bytes, confirm=confirm), "Waiting for image state write...")  # type: ignore # noqa

        if error(r):
            print(r)
        elif success(r):
            pass
        else:
            raise Exception("Unreachable")

    asyncio.run(f())


async def upload_with_progress_bar(
    smpclient: SMPClient, file: typer.FileBinaryRead | BufferedReader, slot: int = 0
) -> None:
    """Animate a progress bar while uploading the FW image."""

    with Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    ) as progress:
        image = file.read()
        file.close()
        task = progress.add_task("Uploading", total=len(image), filename=file.name, start=True)
        try:
            async for offset in smpclient.upload(image, slot):
                progress.update(task, completed=offset)
                logger.info(f"Upload {offset=}")
        except SMPBadStartDelimiter as e:
            progress.stop()
            logger.info(f"Bad start delimiter: {e}")
            logger.error("Got an unexpected response, is the device an SMP server?")
            raise typer.Exit(code=1)
        except OSError as e:
            logger.error(f"Connection to device lost: {e.__class__.__name__} - {e}")
            raise typer.Exit(code=1)


@app.command()
def upload(
    ctx: typer.Context,
    file: Annotated[Path, typer.Argument(help="Path to FW image")],
    slot: Annotated[int, typer.Option(help="The image slot to upload to")] = 0,
) -> None:
    """Upload a FW image."""

    try:
        image_info = ImageInfo.load_file(str(file))
        logger.info(str(image_info))
    except Exception:
        logger.exception("Inspection of FW image failed")
        raise typer.Exit(code=1)

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)
        with open(file, "rb") as f:
            await upload_with_progress_bar(smpclient, f, slot)

    asyncio.run(f())



"""The Intercreate (ic) subcommand group."""

import asyncio
import logging
from io import BufferedReader
from pathlib import Path
from typing import cast

import typer
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from smp import header as smphdr
from smp.exceptions import SMPBadStartDelimiter
from smpclient.extensions import intercreate as ic
from typing_extensions import Annotated

from smpmgr.common import Options, connect_with_spinner, get_custom_smpclient

app = typer.Typer(
    name="ic", help=f"The Intercreate User Group ({smphdr.UserGroupId.INTERCREATE.value})"
)
logger = logging.getLogger(__name__)


async def upload_with_progress_bar(
    smpclient: ic.ICUploadClient, file: typer.FileBinaryRead | BufferedReader, image: int = 0
) -> None:
    """Animate a progress bar while uploading the data."""

    with Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    ) as progress:
        data = file.read()
        file.close()
        task = progress.add_task("Uploading", total=len(data), filename=file.name, start=True)
        try:
            async for offset in smpclient.ic_upload(data, image):
                progress.update(task, completed=offset)
                logger.info(f"Upload {offset=}")
        except SMPBadStartDelimiter as e:
            progress.stop()
            logger.info(f"Bad start delimiter: {e}")
            logger.error("Got an unexpected response, is the device an SMP server?")
            raise typer.Exit(code=1)
        except OSError as e:
            logger.error(f"Connection to device lost: {e.__class__.__name__} - {e}")
            raise typer.Exit(code=1)


@app.command()
def upload(
    ctx: typer.Context,
    file: Annotated[Path, typer.Argument(help="Path to binary data")],
    image: Annotated[int, typer.Option(help="The image slot to upload to")] = 0,
) -> None:
    """Upload data to custom image slots, like a secondary MCU or external storage."""

    options = cast(Options, ctx.obj)
    smpclient = get_custom_smpclient(options, ic.ICUploadClient)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)
        with open(file, "rb") as f:
            await upload_with_progress_bar(smpclient, f, image)

    asyncio.run(f())



"""Runtime discovery and execution of user-provided plugins."""

import logging
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Final, NamedTuple

import typer

logger: Final = logging.getLogger(__name__)


class Plugin(NamedTuple):
    """An SMP group plugin."""

    app: typer.Typer
    path: Path


def get_plugins(argv: list[str]) -> tuple[Plugin, ...]:
    """Returns a tuple of plugin paths and removes them from `argv`."""

    logger.debug(f"{argv=}")
    paths: Final[list[Path]] = []
    for arg in argv[:]:
        if arg.startswith("--plugin-path="):
            logger.debug(f"Found plugin path argument: {arg}")
            path = Path(arg.split("=")[1]).expanduser().resolve()
            if path.is_dir():
                logger.debug(f"Adding plugin path and removing from argv: {path}")
                paths.append(path)
                argv.remove(arg)
            else:
                raise ValueError(f"Invalid plugin path: {path}")

    files: Final = tuple(file for path in paths for file in path.glob("*_group.py"))

    plugins: Final[list[Plugin]] = []
    for file in files:
        if (spec := spec_from_file_location(file.stem, file)) and spec.loader:
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
        else:
            raise ImportError(f"Could not load module from {file}")
        app = getattr(module, "app", None)
        if app is None:
            raise ImportError(f"Module {file} does not have an 'app' attribute")
        if not isinstance(app, typer.Typer):
            raise TypeError(f"Module {file} 'app' attribute is not a typer.Typer")
        plugins.append(Plugin(app=app, path=file))

    logger.debug(f"Plugins found: {plugins}")

    return tuple(plugins)



"""Entry point for the `smpmgr` application."""

import asyncio
import logging
import sys
from importlib.metadata import version as get_version
from pathlib import Path
from typing import Final, cast

import typer
import typer.rich_utils
from rich import print
from smp import error as smperr
from smp.os_management import OS_MGMT_RET_RC
from smpclient.generics import error, error_v1, error_v2, success
from smpclient.mcuboot import IMAGE_TLV, ImageInfo, TLVNotFound
from smpclient.requests.image_management import ImageStatesWrite
from smpclient.requests.os_management import ResetWrite
from typing_extensions import Annotated, assert_never

from smpmgr import file_management, image_management, os_management, terminal
from smpmgr.common import (
    Options,
    TransportDefinition,
    connect_with_spinner,
    get_smpclient,
    smp_request,
)
from smpmgr.image_management import upload_with_progress_bar
from smpmgr.logging import LogLevel, setup_logging
from smpmgr.plugins import get_plugins
from smpmgr.user import intercreate

logger = logging.getLogger(__name__)

# Intercept and modify sys.argv to look for plugins
plugins: Final = get_plugins(sys.argv)

HELP_LINES: Final = (
    f"Simple Management Protocol (SMP) Manager Version {get_version('smpmgr')}\n",
    "\n[dim]Copyright (c) 2023-2025 Intercreate, Inc. and Contributors[/dim]\n",
) + (
    (
        "[bold yellow]"
        "\nNOTE: Plugins have been loaded from the following source(s). Ensure that you "
        "trust the files that have been loaded. The developers and contributors of "
        "this application shall not be held liable for any damages, losses, or other "
        "consequences arising from the use or misuse of this application.\n\n"
        "[/bold yellow]",
        "\n".join(f"  - {plugin.path.resolve()}" for plugin in plugins),
    )
    if plugins
    else ()
)

# Override the dimming of the help text
typer.rich_utils.STYLE_HELPTEXT = ""

app: Final = typer.Typer(help="".join(HELP_LINES), rich_markup_mode="rich")
app.add_typer(os_management.app)
app.add_typer(image_management.app)
app.add_typer(file_management.app)
app.add_typer(intercreate.app)
app.command()(terminal.terminal)

for plugin in plugins:
    app.add_typer(plugin.app)


@app.callback(invoke_without_command=True)
def options(
    ctx: typer.Context,
    ip: str = typer.Option(None, help="The IP address to connect to for UDP transport"),
    port: str = typer.Option(
        None, help="The serial port to connect to, e.g. COM1, /dev/ttyACM0, etc."
    ),
    ble: str = typer.Option(None, help="The Bluetooth address to connect to"),
    timeout: float = typer.Option(
        2.0, help="Transport timeout in seconds; how long to wait for requests"
    ),
    mtu: int = typer.Option(
        4096,
        help=(
            "Maximum transmission unit supported by the SMP server serial transport."
            "  Ignored for BLE transport since the BLE connection will report MTU."
        ),
    ),
    loglevel: LogLevel = typer.Option(None, help="Debug log level"),
    logfile: Path = typer.Option(None, help="Log file path"),
    version: Annotated[bool, typer.Option("--version", help="Show the version and exit.")] = False,
    plugin_path: Path = typer.Option(
        None, help="Path to plugin directory. May be used more than once."
    ),
) -> None:
    if plugin_path:
        raise ValueError(
            "--plugin-path should have been removed from sys.argv by the get_plugins(sys.argv) "
            f"call. This is a bug! {sys.argv=}"
        )
    if version:
        print(get_version('smpmgr'))
        raise typer.Exit()

    setup_logging(loglevel, logfile)

    ctx.obj = Options(
        timeout=timeout, transport=TransportDefinition(port=port, ble=ble, ip=ip), mtu=mtu
    )
    logger.info(ctx.obj)

    if ctx.invoked_subcommand is None:
        if loglevel is not None or logfile is not None:
            raise typer.Exit()
        print("A command is required, see [bold]--help[/bold] for available commands.")
        raise typer.Exit()

    # TODO: type of transport is inferred from the argument given (--port, --ble, --usb, etc), but
    # it must be the case that only one is provided.


@app.command()
def upgrade(
    ctx: typer.Context,
    file: Annotated[Path, typer.Argument(help="Path to FW image")],
    slot: Annotated[int, typer.Option(help="The image slot to upload to")] = 0,
) -> None:
    """Upload a FW image, mark it for next boot, and reset the device."""

    try:
        image_info = ImageInfo.load_file(str(file))
        logger.info(str(image_info))
    except Exception as e:
        typer.echo(f"Inspection of FW image failed: {e}")
        raise typer.Exit(code=1)

    try:
        image_tlv_sha256 = image_info.get_tlv(IMAGE_TLV.SHA256)
        logger.info(f"IMAGE_TLV_SHA256: {image_tlv_sha256}")
    except TLVNotFound:
        typer.echo("Could not find IMAGE_TLV_SHA256 in image.")
        raise typer.Exit(code=1)

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)

        with open(file, "rb") as f:
            await upload_with_progress_bar(smpclient, f, slot)

        if slot != 0:
            # mark the new image for testing (swap)
            image_states_response = await smp_request(
                smpclient,
                options,
                ImageStatesWrite(hash=image_tlv_sha256.value),
                "Marking uploaded image for test upgrade...",
            )
            if success(image_states_response):
                pass
            elif error(image_states_response):
                print(image_states_response)
                raise typer.Exit(code=1)
            else:
                assert_never(image_states_response)

        reset_response = await smp_request(smpclient, options, ResetWrite())
        if success(reset_response):
            pass
        elif error(reset_response):
            if error_v1(reset_response):
                if reset_response.rc != smperr.MGMT_ERR.EOK:
                    print(reset_response)
                    raise typer.Exit(code=1)
            elif error_v2(reset_response):
                if reset_response.err.rc != OS_MGMT_RET_RC.OK:
                    print(reset_response)
                    raise typer.Exit(code=1)
            else:
                assert_never(reset_response)
        else:
            assert_never(reset_response)

        print("Upgrade complete.")

        if slot != 0:
            print("The device may take a few minutes to complete FW swap.")

    asyncio.run(f())


@app.command()
def shell() -> None:
    """Open the `smpmgr` interactive shell. Type 'exit' or 'quit' to exit."""

    print("".join(HELP_LINES))
    print("Type 'exit' or 'quit' to exit the shell.\n")

    while True:
        args = typer.prompt("smpmgr", prompt_suffix=' >').split()

        if args[0] in {"exit", "quit"}:
            break
        if args[0] == "shell":
            print("The 'shell' command cannot be used from within the shell.")
            continue

        try:
            app(args)
        except SystemExit:
            continue



"""Common CLI helpers from rich, typer, click, etc."""

import asyncio
import logging
from dataclasses import dataclass, fields
from typing import Type, TypeVar

import typer
from rich.progress import Progress, SpinnerColumn, TextColumn
from serial import SerialException
from smp.exceptions import SMPBadStartDelimiter
from smpclient import SMPClient
from smpclient.generics import SMPRequest, TEr1, TEr2, TRep
from smpclient.transport.ble import SMPBLETransport
from smpclient.transport.serial import SMPSerialTransport
from smpclient.transport.udp import SMPUDPTransport

logger = logging.getLogger(__name__)

TSMPClient = TypeVar(
    "TSMPClient",
    bound=SMPClient,
)


@dataclass(frozen=True)
class TransportDefinition:
    port: str | None
    ble: str | None
    ip: str | None


@dataclass(frozen=True)
class Options:
    timeout: float
    transport: TransportDefinition
    mtu: int


def get_custom_smpclient(options: Options, smp_client_cls: Type[TSMPClient]) -> TSMPClient:
    """Return an `SMPClient` subclass to the chosen transport or raise `typer.Exit`."""
    if options.transport.port is not None:
        logger.info(
            f"Initializing SMPClient with the SMPSerialTransport, {options.transport.port=}"
        )
        if options.mtu is not None:
            return smp_client_cls(
                SMPSerialTransport(
                    max_smp_encoded_frame_size=options.mtu, line_length=options.mtu, line_buffers=1
                ),
                options.transport.port,
            )
        else:
            return smp_client_cls(SMPSerialTransport(), options.transport.port)
    elif options.transport.ble is not None:
        logger.info(f"Initializing SMPClient with the SMPBLETransport, {options.transport.ble=}")
        return smp_client_cls(
            SMPBLETransport(),
            options.transport.ble,
        )
    elif options.transport.ip is not None:
        logger.info(
            f"Initializing SMPClient with the SMPUDPTransport, {options.transport.ip=} and MTU 1024"
        )
        return smp_client_cls(
            SMPUDPTransport(1024),
            options.transport.ip,
        )
    else:
        typer.echo(
            f"A transport option is required; "
            f"one of [{', '.join(map(lambda x: '--' + x.name, fields(options.transport)))}]."
        )
        typer.echo("See smpmgr --help.")
        raise typer.Exit(code=1)


def get_smpclient(options: Options) -> SMPClient:
    """Return an `SMPClient` to the chosen transport or raise `typer.Exit`."""
    return get_custom_smpclient(options, SMPClient)


async def connect_with_spinner(smpclient: SMPClient, timeout_s: float) -> None:
    """Spin while connecting to the SMP Server; raises `typer.Exit` if connection fails."""
    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}")
    ) as progress:
        connect_task_description = f"Connecting to {smpclient._address}..."
        connect_task = progress.add_task(description=connect_task_description, total=None)
        try:
            await asyncio.wait_for(smpclient.connect(), timeout=timeout_s)
            progress.update(
                connect_task, description=f"{connect_task_description} OK", completed=True
            )
            return
        except asyncio.TimeoutError:
            logger.error("Transport error: connection timeout")
        except SerialException as e:
            logger.error(f"Serial transport error: {e.__class__.__name__} - {e}")

        progress.update(
            connect_task, description=f"{connect_task_description} error", completed=True
        )
        raise typer.Exit(code=1)


async def smp_request(
    smpclient: SMPClient,
    options: Options,
    request: SMPRequest[TRep, TEr1, TEr2],
    description: str | None = None,
) -> TRep | TEr1 | TEr2:
    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}")
    ) as progress:
        description = description or f"Waiting for response to {request.__class__.__name__}..."
        task = progress.add_task(description=description, total=None)
        try:
            r = await asyncio.wait_for(smpclient.request(request), timeout=options.timeout)
            progress.update(task, description=f"{description} OK", completed=True)
            return r
        except asyncio.TimeoutError:
            progress.update(task, description=f"{description} timeout", completed=True)
            logger.error("Timeout waiting for response")
            raise typer.Exit(code=1)
        except SMPBadStartDelimiter:
            progress.update(task, description=f"{description} SMP error", completed=True)
            logger.error("Is the device an SMP server?")
            raise typer.Exit(code=1)
        except OSError as e:
            progress.update(task, description=f"{description} OS error", completed=True)
            logger.error(f"Connection to device lost: {e.__class__.__name__} - {e}")
            raise typer.Exit(code=1)



"""Logging configuration for smpmgr."""

import asyncio
import logging
from enum import Enum, unique
from pathlib import Path
from typing import Dict

import click
import serial
import typer
from rich.logging import RichHandler


@unique
class LogLevel(Enum):
    CRITICAL = 'CRITICAL'
    ERROR = 'ERROR'
    WARNING = 'WARNING'
    INFO = 'INFO'
    DEBUG = 'DEBUG'
    NOTSET = 'NOTSET'


def setup_logging(loglevel: LogLevel | None, logfile: Path | None) -> None:
    """Setup logging for smpmgr.

    During normal operation, this function is called once at the start of the program.  However,
    when using the `shell` command, this function may be called multiple times.  This allows the
    user to change the log level and log file without restarting the program.

    For example, to change the log level to DEBUG and the log file to `smpmgr.log`:

    ```shell
    smpmgr > --loglevel DEBUG --logfile smpmgr.log
    ```
    This is a valid stand alone command that will save those log settings for the duration of the
    shell session, until one or the other is updated with new values.
    """

    DEBUG_FORMAT = "%(message)s - %(pathname)s:%(lineno)s"
    DEFAULT_FORMAT = "%(message)s - %(module)s:%(lineno)s"
    LOGFILE_FORMAT = "%(asctime)s - %(levelname)s - %(pathname)s:%(lineno)s - %(message)s"
    NAME_CONSOLE_HANDLER = "console_handler"
    NAME_FILE_HANDLER = "file_handler"

    base_console_handler = RichHandler(
        rich_tracebacks=True, tracebacks_suppress=[click, typer, asyncio, serial]
    )
    base_console_handler.name = NAME_CONSOLE_HANDLER

    console_handler = (
        RichHandler(rich_tracebacks=True, tracebacks_suppress=[click, typer, asyncio, serial])
        if loglevel is not None
        else None
    )
    if console_handler is not None:
        console_handler.name = NAME_CONSOLE_HANDLER

    file_handler = logging.FileHandler(logfile) if logfile is not None else None
    if file_handler is not None:
        file_handler.name = NAME_FILE_HANDLER

    # create a map of handler names -> handlers
    new_handlers: Dict[str, logging.Handler] = {
        h.name: h  # type: ignore
        for h in [
            console_handler,
            file_handler,
        ]
        if h is not None
    }

    # get a map of existing handler names -> handlers
    old_handlers: Dict[str, logging.Handler] = {h.name: h for h in logging.root.handlers}  # type: ignore # noqa: E501

    # update the old handlers with the new handlers
    handlers = {base_console_handler.name: base_console_handler} | old_handlers | new_handlers

    logging.basicConfig(
        level=logging.NOTSET,  # root logger logs everything
        format=(DEBUG_FORMAT if loglevel == LogLevel.DEBUG else DEFAULT_FORMAT),
        datefmt="[%X]",
        handlers=handlers.values(),
        force=True,
    )

    handlers[NAME_CONSOLE_HANDLER].setLevel(
        loglevel.value if loglevel is not None else logging.WARNING
    )  # UI console log level set from --loglevel
    logging.info(f"Console log level: {logging.getLevelName(handlers[NAME_CONSOLE_HANDLER].level)}")

    if file_handler is not None:
        file_handler.setLevel(logging.DEBUG)  # file logs are always DEBUG
        file_handler.setFormatter(logging.Formatter(LOGFILE_FORMAT))
        logging.info(f"Log file {logfile} log level: {logging.getLevelName(file_handler.level)}")



"""The image subcommand group."""

import asyncio
import logging
from io import BufferedReader
from pathlib import Path
from typing import cast

import typer
from rich import print
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from smp.exceptions import SMPBadStartDelimiter
from smpclient import SMPClient
from smpclient.generics import error, success
from smpclient.requests.file_management import (
    FileHashChecksum,
    FileStatus,
    SupportedFileHashChecksumTypes,
)
from typing_extensions import Annotated

from smpmgr.common import Options, connect_with_spinner, get_smpclient, smp_request

app = typer.Typer(name="file", help="The SMP File Management Group.")
logger = logging.getLogger(__name__)


@app.command()
def get_supported_hash_types(ctx: typer.Context) -> None:
    """Request the supported hash types."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)

        r = await smp_request(smpclient, options, SupportedFileHashChecksumTypes(), "Waiting for supported hash types...")  # type: ignore # noqa

        if error(r):
            print(r)
        elif success(r):
            print(r.types)
        else:
            raise Exception("Unreachable")

    asyncio.run(f())


@app.command()
def get_hash(
    ctx: typer.Context, file: Annotated[str, typer.Argument(help="Path to file on the SMP Server")]
) -> None:
    """Request the hash of a file on the SMP Server."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)

        r = await smp_request(smpclient, options, FileHashChecksum(name=file), "Waiting for hash...")  # type: ignore # noqa

        if error(r) or success(r):
            print(r)
        else:
            raise Exception("Unreachable")

    asyncio.run(f())


@app.command()
def read_size(
    ctx: typer.Context, file: Annotated[str, typer.Argument(help="Path to file on the SMP Server")]
) -> None:
    """Request to read the size of a file on the SMP Server."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)

        r = await smp_request(smpclient, options, FileStatus(name=file), "Waiting for file size...")  # type: ignore # noqa

        if error(r):
            print(r)
        elif success(r):
            print(r.len)
        else:
            raise Exception("Unreachable")

    asyncio.run(f())


async def upload_with_progress_bar(
    smpclient: SMPClient, file: typer.FileBinaryRead | BufferedReader, destination: str
) -> None:
    """Animate a progress bar while uploading the file."""

    with Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    ) as progress:
        file_data = file.read()
        file.close()
        task = progress.add_task("Uploading", total=len(file_data), filename=file.name, start=True)
        try:
            async for offset in smpclient.upload_file(file_data, destination):
                progress.update(task, completed=offset)
                logger.info(f"Upload {offset=}")
        except SMPBadStartDelimiter as e:
            progress.stop()
            logger.info(f"Bad start delimiter: {e}")
            logger.error("Got an unexpected response, is the device an SMP server?")
            raise typer.Exit(code=1)
        except OSError as e:
            logger.error(f"Connection to device lost: {e.__class__.__name__} - {e}")
            raise typer.Exit(code=1)


@app.command()
def upload(
    ctx: typer.Context,
    file: Annotated[Path, typer.Argument(help="Path to file")],
    destination: Annotated[str, typer.Argument(help="The destination on the SMP Server")],
) -> None:
    """Upload a file."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)
        with open(file, "rb") as f:
            await upload_with_progress_bar(smpclient, f, destination)

    asyncio.run(f())


@app.command()
def download(
    ctx: typer.Context,
    file: Annotated[str, typer.Argument(help="The file on the SMP Server")],
    destination: Annotated[Path, typer.Argument(help="The destination on the local filesystem")],
) -> None:
    """Download a file."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)
        with destination.open("wb") as dest_f:
            file_data = await smpclient.download_file(file)
            dest_f.write(file_data)

    asyncio.run(f())



import asyncio
from typing import cast

import typer
from rich import print
from smpclient.requests.os_management import EchoWrite, ResetWrite

from smpmgr.common import Options, connect_with_spinner, get_smpclient, smp_request

app = typer.Typer(name="os", help="The SMP OS Management Group.")


@app.command()
def echo(ctx: typer.Context, message: str) -> None:
    """Request that the SMP Server echo the given message."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)
        r = await smp_request(smpclient, options, EchoWrite(d=message))  # type: ignore
        print(r)

    asyncio.run(f())


@app.command()
def reset(ctx: typer.Context) -> None:
    """Request that the SMP Server reset the device."""

    options = cast(Options, ctx.obj)
    smpclient = get_smpclient(options)

    async def f() -> None:
        await connect_with_spinner(smpclient, options.timeout)
        r = await smp_request(smpclient, options, ResetWrite())  # type: ignore
        print(r)

    asyncio.run(f())
